@@ -1,14 +1,11 @@
-from ..utils import AttrDict, AttrList
+from ..utils import AttrDict, AttrList, _wrap
 
 from .hit import Hit, HitMeta
 
-class SuggestResponse(AttrDict):
-    def success(self):
-        return not self._shards.failed
-
 class Response(AttrDict):
-    def __init__(self, search, response):
+    def __init__(self, search, response, doc_class=None):
         super(AttrDict, self).__setattr__('_search', search)
+        super(AttrDict, self).__setattr__('_doc_class', doc_class)
         super(Response, self).__init__(response)
 
     def __iter__(self):
@@ -25,36 +22,29 @@ class Response(AttrDict):
     __bool__ = __nonzero__
 
     def __repr__(self):
-        return '<Response: %r>' % self.hits
+        return '<Response: %r>' % (self.hits or self.aggregations)
 
     def __len__(self):
         return len(self.hits)
 
     def __getstate__(self):
-        return (self._d_, self._search)
+        return self._d_, self._search, self._doc_class
 
     def __setstate__(self, state):
         super(AttrDict, self).__setattr__('_d_', state[0])
         super(AttrDict, self).__setattr__('_search', state[1])
+        super(AttrDict, self).__setattr__('_doc_class', state[2])
 
     def success(self):
         return self._shards.total == self._shards.successful and not self.timed_out
 
-    def _get_result(self, hit):
-        dt = hit.get('_type')
-        for t in hit.get('inner_hits', ()):
-            hit['inner_hits'][t] = Response(self._search, hit['inner_hits'][t])
-        callback = self._search._doc_type_map.get(dt, Hit)
-        callback = getattr(callback, 'from_es', callback)
-        return callback(hit)
-
     @property
     def hits(self):
         if not hasattr(self, '_hits'):
             h = self._d_['hits']
 
             try:
-                hits = AttrList(map(self._get_result, h['hits']))
+                hits = AttrList(map(self._search._get_result, h['hits']))
             except AttributeError as e:
                 # avoid raising AttributeError since it will be hidden by the property
                 raise TypeError("Could not parse hits.", e)
@@ -62,7 +52,7 @@ class Response(AttrDict):
             # avoid assigning _hits into self._d_
             super(AttrDict, self).__setattr__('_hits', hits)
             for k in h:
-                setattr(self._hits, k, h[k])
+                setattr(self._hits, k, _wrap(h[k]))
         return self._hits
 
     @property
@@ -94,3 +84,10 @@ class AggResponse(AttrDict):
         for name in self._meta['aggs']:
             yield self[name]
 
+
+class UpdateByQueryResponse(AttrDict):
+
+    def __init__(self, search, response, doc_class=None):
+        super(AttrDict, self).__setattr__('_search', search)
+        super(AttrDict, self).__setattr__('_doc_class', doc_class)
+        super(UpdateByQueryResponse, self).__init__(response)
