@@ -1,16 +1,22 @@
 import copy
-import collections
+
+try:
+    import collections.abc as collections_abc  # only works on python 3.3+
+except ImportError:
+    import collections as collections_abc
 
 from six import iteritems, string_types
 
 from elasticsearch.helpers import scan
 from elasticsearch.exceptions import TransportError
 
-from .query import Q, EMPTY_QUERY, Bool
+from .query import Q, Bool
 from .aggs import A, AggBase
 from .utils import DslBase, AttrDict
-from .response import Response, Hit, SuggestResponse
+from .response import Response, Hit
 from .connections import connections
+from .exceptions import IllegalOperation
+
 
 class QueryProxy(object):
     """
@@ -20,16 +26,23 @@ class QueryProxy(object):
     """
     def __init__(self, search, attr_name):
         self._search = search
-        self._proxied = EMPTY_QUERY
+        self._proxied = None
         self._attr_name = attr_name
 
     def __nonzero__(self):
-        return self._proxied != EMPTY_QUERY
+        return self._proxied is not None
     __bool__ = __nonzero__
 
     def __call__(self, *args, **kwargs):
         s = self._search._clone()
-        getattr(s, self._attr_name)._proxied &= Q(*args, **kwargs)
+
+        # we cannot use self._proxied since we just cloned self._search and
+        # need to access the new self on the clone
+        proxied = getattr(s, self._attr_name)
+        if proxied._proxied is None:
+            proxied._proxied = Q(*args, **kwargs)
+        else:
+            proxied._proxied &= Q(*args, **kwargs)
 
         # always return search to be chainable
         return s
@@ -44,7 +57,7 @@ class QueryProxy(object):
         super(QueryProxy, self).__setattr__(attr_name, value)
 
     def __getstate__(self):
-        return (self._search, self._proxied, self._attr_name)
+        return self._search, self._proxied, self._attr_name
 
     def __setstate__(self, state):
         self._search, self._proxied, self._attr_name = state
@@ -72,12 +85,14 @@ class ProxyDescriptor(object):
 class AggsProxy(AggBase, DslBase):
     name = 'aggs'
     def __init__(self, search):
-        self._base = self._search = search
+        self._base = self
+        self._search = search
         self._params = {'aggs': {}}
 
     def to_dict(self):
         return super(AggsProxy, self).to_dict().get('aggs', {})
 
+
 class Request(object):
     def __init__(self, using='default', index=None, doc_type=None, extra=None):
         self._using = using
@@ -91,13 +106,12 @@ class Request(object):
         self._doc_type = []
         self._doc_type_map = {}
         if isinstance(doc_type, (tuple, list)):
-            for dt in doc_type:
-                self._add_doc_type(dt)
-        elif isinstance(doc_type, collections.Mapping):
+            self._doc_type.extend(doc_type)
+        elif isinstance(doc_type, collections_abc.Mapping):
             self._doc_type.extend(doc_type.keys())
             self._doc_type_map.update(doc_type)
         elif doc_type:
-            self._add_doc_type(doc_type)
+            self._doc_type.append(doc_type)
 
         self._params = {}
         self._extra = extra or {}
@@ -111,6 +125,9 @@ class Request(object):
             other.to_dict() == self.to_dict()
         )
 
+    def __copy__(self):
+        return self._clone()
+
     def params(self, **kwargs):
         """
         Specify query params to be used when executing the search. All the
@@ -135,25 +152,80 @@ class Request(object):
 
             s = Search()
             s = s.index('twitter-2015.01.01', 'twitter-2015.01.02')
+            s = s.index(['twitter-2015.01.01', 'twitter-2015.01.02'])
         """
         # .index() resets
         s = self._clone()
         if not index:
             s._index = None
         else:
-            s._index = (self._index or []) + list(index)
+            indexes = []
+            for i in index:
+                if isinstance(i, string_types):
+                    indexes.append(i)
+                elif isinstance(i, list):
+                    indexes += i
+                elif isinstance(i, tuple):
+                    indexes += list(i)
+
+            s._index = (self._index or []) + indexes
+
         return s
 
-    def _add_doc_type(self, doc_type):
-        if hasattr(doc_type, '_doc_type'):
-            self._doc_type_map[doc_type._doc_type.name] = doc_type
-            doc_type = doc_type._doc_type.name
-        self._doc_type.append(doc_type)
+    def _resolve_field(self, path):
+        for dt in self._doc_type:
+            if not hasattr(dt, '_index'):
+                continue
+            field = dt._index.resolve_field(path)
+            if field is not None:
+                return field
+
+    def _resolve_nested(self, hit, parent_class=None):
+        doc_class = Hit
+
+        nested_path = []
+        nesting = hit['_nested']
+        while nesting and 'field' in nesting:
+            nested_path.append(nesting['field'])
+            nesting = nesting.get('_nested')
+        nested_path = '.'.join(nested_path)
+
+        if hasattr(parent_class, '_index'):
+            nested_field = parent_class._index.resolve_field(nested_path)
+        else:
+            nested_field = self._resolve_field(nested_path)
+
+        if nested_field is not None:
+            return nested_field._doc_class
+
+        return doc_class
+
+    def _get_result(self, hit, parent_class=None):
+        doc_class = Hit
+        dt = hit.get('_type')
+
+        if '_nested' in hit:
+            doc_class = self._resolve_nested(hit, parent_class)
+
+        elif dt in self._doc_type_map:
+            doc_class = self._doc_type_map[dt]
+
+        else:
+            for doc_type in self._doc_type:
+                if hasattr(doc_type, '_matches') and doc_type._matches(hit):
+                    doc_class = doc_type
+                    break
+
+        for t in hit.get('inner_hits', ()):
+            hit['inner_hits'][t] = Response(self, hit['inner_hits'][t], doc_class=doc_class)
+
+        callback = getattr(doc_class, 'from_es', doc_class)
+        return callback(hit)
 
     def doc_type(self, *doc_type, **kwargs):
         """
         Set the type to search through. You can supply a single value or
-        multiple. Values can be strings or subclasses of ``DocType``.
+        multiple. Values can be strings or subclasses of ``Document``.
 
         You can also pass in any keyword arguments, mapping a doc_type to a
         callback that should be used instead of the Hit class.
@@ -171,8 +243,7 @@ class Request(object):
             s._doc_type = []
             s._doc_type_map = {}
         else:
-            for dt in doc_type:
-                s._add_doc_type(dt)
+            s._doc_type.extend(doc_type)
             s._doc_type.extend(kwargs.keys())
             s._doc_type_map.update(kwargs)
         return s
@@ -223,7 +294,7 @@ class Search(Request):
         :arg doc_type: only query this type.
 
         All the parameters supplied (or omitted) at creation type can be later
-        overriden by methods (`using`, `index` and `doc_type` respectively).
+        overridden by methods (`using`, `index` and `doc_type` respectively).
         """
         super(Search, self).__init__(**kwargs)
 
@@ -351,8 +422,8 @@ class Search(Request):
         aggs = d.pop('aggs', d.pop('aggregations', {}))
         if aggs:
             self.aggs._params = {
-                'aggs': dict(
-                    (name, A(value)) for (name, value) in iteritems(aggs))
+                'aggs': {
+                    name: A(value) for (name, value) in iteritems(aggs)}
             }
         if 'sort' in d:
             self._sort = d.pop('sort')
@@ -385,8 +456,10 @@ class Search(Request):
             s = s.script_fields(times_two="doc['field'].value * 2")
             s = s.script_fields(
                 times_three={
-                    'script': "doc['field'].value * n",
-                    'params': {'n': 3}
+                    'script': {
+                        'inline': "doc['field'].value * params.n",
+                        'params': {'n': 3}
+                    }
                 }
             )
 
@@ -402,10 +475,10 @@ class Search(Request):
         """
         Selectively control how the _source field is returned.
 
-        :arg source: wildcard string, array of wildcards, or dictionary of includes and excludes
+        :arg fields: wildcard string, array of wildcards, or dictionary of includes and excludes
 
-        If ``source`` is None, the entire document will be returned for
-        each hit.  If source is a dictionary with keys of 'include' and/or
+        If ``fields`` is None, the entire document will be returned for
+        each hit.  If fields is a dictionary with keys of 'include' and/or
         'exclude' the fields will be either included or excluded appropriately.
 
         Calling this multiple times with the same named parameter will override the
@@ -470,6 +543,8 @@ class Search(Request):
         s._sort = []
         for k in keys:
             if isinstance(k, string_types) and k.startswith('-'):
+                if k[1:] == '_score':
+                    raise IllegalOperation('Sorting by `-_score` is not allowed.')
                 k = {k[1:]: {"order": "desc"}}
             s._sort.append(k)
         return s
@@ -489,7 +564,7 @@ class Search(Request):
     def highlight(self, *fields, **kwargs):
         """
         Request highlighting of some fields. All keyword arguments passed in will be
-        used as parameters. Example::
+        used as parameters for all the fields in the ``fields`` parameter. Example::
 
             Search().highlight('title', 'body', fragment_size=50)
 
@@ -504,6 +579,21 @@ class Search(Request):
                 }
             }
 
+        If you want to have different options for different fields you can call ``highlight`` twice::
+
+            Search().highlight('title', fragment_size=50).highlight('body', fragment_size=100)
+
+        which will produce::
+
+            {
+                "highlight": {
+                    "fields": {
+                        "body": {"fragment_size": 100},
+                        "title": {"fragment_size": 50}
+                    }
+                }
+            }
+
         """
         s = self._clone()
         for f in fields:
@@ -532,12 +622,15 @@ class Search(Request):
         Serialize the search into the dictionary that will be sent over as the
         request's body.
 
-        :arg count: a flag to specify we are interested in a body for count -
+        :arg count: a flag to specify if we are interested in a body for count -
             no aggregations, no pagination bounds etc.
 
         All additional keyword arguments will be included into the dictionary.
         """
-        d = {"query": self.query.to_dict()}
+        d = {}
+
+        if self.query:
+            d["query"] = self.query.to_dict()
 
         # count request doesn't care for sorting and other things
         if not count:
@@ -552,7 +645,7 @@ class Search(Request):
 
             d.update(self._extra)
 
-            if not self._source in (None, {}):
+            if self._source not in (None, {}):
                 d['_source'] = self._source
 
             if self._highlight:
@@ -582,7 +675,6 @@ class Search(Request):
         # TODO: failed shards detection
         return es.count(
             index=self._index,
-            doc_type=self._doc_type,
             body=d,
             **self._params
         )['count']
@@ -592,7 +684,8 @@ class Search(Request):
         Execute the search and return an instance of ``Response`` wrapping all
         the data.
 
-        :arg response_class: optional subclass of ``Response`` to use instead.
+        :arg ignore_cache: if set to ``True``, consecutive calls will hit
+            ES, while cached result will be ignored. Defaults to `False`
         """
         if ignore_cache or not hasattr(self, '_response'):
             es = connections.get_connection(self._using)
@@ -601,27 +694,12 @@ class Search(Request):
                 self,
                 es.search(
                     index=self._index,
-                    doc_type=self._doc_type,
                     body=self.to_dict(),
                     **self._params
                 )
             )
         return self._response
 
-    def execute_suggest(self):
-        """
-        Execute just the suggesters. Ignores all parts of the request that are
-        not relevant, including ``query`` and ``doc_type``.
-        """
-        es = connections.get_connection(self._using)
-        return SuggestResponse(
-            es.suggest(
-                index=self._index,
-                body=self._suggest,
-                **self._params
-            )
-        )
-
     def scan(self):
         """
         Turn the search into a scan search and return a generator that will
@@ -638,12 +716,9 @@ class Search(Request):
                 es,
                 query=self.to_dict(),
                 index=self._index,
-                doc_type=self._doc_type,
                 **self._params
-            ):
-            callback = self._doc_type_map.get(hit['_type'], Hit)
-            callback = getattr(callback, 'from_es', callback)
-            yield callback(hit)
+        ):
+            yield self._get_result(hit)
 
     def delete(self):
         """
@@ -656,13 +731,11 @@ class Search(Request):
             es.delete_by_query(
                 index=self._index,
                 body=self.to_dict(),
-                doc_type=self._doc_type,
                 **self._params
             )
         )
 
 
-
 class MultiSearch(Request):
     """
     Combine multiple :class:`~elasticsearch_dsl.Search` objects into a single
@@ -701,8 +774,6 @@ class MultiSearch(Request):
             meta = {}
             if s._index:
                 meta['index'] = s._index
-            if s._doc_type:
-                meta['type'] = s._doc_type
             meta.update(s._params)
 
             out.append(meta)
@@ -719,7 +790,6 @@ class MultiSearch(Request):
 
             responses = es.msearch(
                 index=self._index,
-                doc_type=self._doc_type,
                 body=self.to_dict(),
                 **self._params
             )
